// Stat Bar Overlays Library
// Version 0.6 - 2013-10-05
// Create by DarkCampainger

// This library allows you to quickly create stat bar overlays
//   for health, mana, experience, and more. Tell it which
//   variables to track when it's created, and then all you have to
//   do is call Update()

// See included Reference.dm for process reference
// See included demo for example of usage.

bar_overlay
	var
		var_stat // The name of the stat variable (in text) that is being tracked
		var_max // The name of the stat max variable (in text) that is being tracked, or a value to use as the max
		icon // The icon file to use
		state_prefix = "" // Prefix to apply to icon_states (in case of multiple bars in one icon file)
		states // How many different states, or levels. They should be numbered 1,2,3,4,ect in the icon.
		layer = FLOAT_LAYER // What layer to display the bar on
		pixel_x = 0 // pixel_x offset of the bar
		pixel_y = 0 // pixel_y offset of the bar
		atom/owner // The atom the bar is attached to
		timeout = 0 // How long after an update the bar will auto-hide itself, 0 for never
		auto_show // If the stat bar will automatically un-hide itself when updated

		// Internal variables, should not be accessed directly:
		image/_image // image object created at runtime to display bar
		_hidden=0 // Tracks whether stat bar is visible or not
		_timeout_id=0 // Used to ensure only the most recent timeout timer is runinng

	New(atom/owner, icon, states, var_stat, var_max, layer, pixel_x, pixel_y, state_prefix, timeout, auto_show, hide)
		// Assign any variables that were passed
		src.owner = owner
		if(icon != null)
			src.icon = icon
		if(states != null)
			src.states = states
		if(var_stat != null)
			src.var_stat = var_stat
		if(var_max != null)
			src.var_max = var_max
		if(layer != null)
			src.layer = layer
		if(pixel_x != null)
			src.pixel_x = pixel_x
		if(pixel_y != null)
			src.pixel_y = pixel_y
		if(state_prefix != null)
			src.state_prefix = state_prefix
		if(timeout != null)
			src.timeout = timeout

		// If auto_show/hide were unspecified, default them to logical values
		src.auto_show = ( (auto_show != null) ? auto_show : (!!src.timeout) )
		src._hidden = ( (hide != null) ? hide : (!!src.timeout) )

		// If the number of states is unset, attempt to determine it
		if(!src.states)
			var/list/state_names = icon_states(src.icon)
			src.states = 0
			while("[src.state_prefix][src.states+1]" in state_names)
				src.states++

		// Check everything has been passed correctly
		if(src.Validate() && !src._hidden)
			// Set the bar and create /image object
			src.Update()

	Del()
		// If the image is set and the owner still exists
		if(src._image && src.owner)
			// Remove the overlay
			src.owner.overlays -= src._image

		..()

	proc/Validate() // Make sure all data is valid

		// Validate owner, icon, and state count
		if(!owner || !istype(owner))
			CRASH("bar_overlay: bad owner '[owner]' [istype(owner,/datum) ? "([owner.type])" : "(non-object)"]!")
		if(!icon || !isicon(icon))
			CRASH("bar_overlay: bad icon '[icon]'!")
		if(!isnum(states) || states<1)
			CRASH("bar_overlay: bad state count of '[states]'!")

		// Validate var_stat to be a variable name
		if(!var_stat || !istext(var_stat))
			CRASH("bar_overlay: var_stat '[var_stat]' must be the name of the min variable in text!")
		if(!(var_stat in owner.vars))
			CRASH("bar_overlay: var_stat '[var_stat]' is not a variable of [owner.type]!")
		if(!isnum(owner.vars[var_stat]))
			CRASH("bar_overlay: var_stat '[var_stat]' must hold a numerical value!")

		// Validate var_max to be a variable name or numerical value
		if(var_max==null || var_max=="")
			CRASH("bar_overlay: var_max '[var_max]' is not set!")
		else if(istext(var_max))
			if(!(var_max in owner.vars))
				CRASH("bar_overlay: var_max '[var_max]' is not a variable of [owner.type]!")
			if(!isnum(owner.vars[var_max]))
				CRASH("bar_overlay: var_stat '[var_stat]' must hold a numerical value!")
		else if(!isnum(var_max))
			CRASH("bar_overlay: var_max '[var_max]' must be the name of the max variable in text or the max value as a number!")

		// Validate timeout
		if(!isnum(timeout))
			CRASH("bar_overlay: timeout '[timeout]' must be a numerical value!")

		// Succesfully validated settings
		return 1

	proc/Update(_unused=null, force_show=null, force_timeout=null) // Update bar on owner

		// Check the owner still exists
		if(!src.owner)
			CRASH("bar_overlay: owner for '[src.var_stat]/[src.var_max]' bar no longer exists!")

		// Calculate new bar value
		var/max = max(src.owner.vars[src.var_max],0.000001) // The larger value, or denominator
		var/min = min(max(src.owner.vars[src.var_stat],0),max) // The smaller value, or numerator
		var/state = "[src.state_prefix][round((src.states-1)*min/max,1)+1]" // Get the percentage and scale it by number of states

		if(!src._image) // If we need to create an image to use
			src._image = image(icon=src.icon, icon_state=state, layer=src.layer, pixel_x=src.pixel_x, pixel_y=src.pixel_y)

		else // Otherwise we can just update it

			src.owner.overlays -= src._image // Need to remove existing overlay before changing image

			src._image.icon_state = state // Update image's icon_state

		// If not hidden, or auto_show/force_show are enabled
		if(!src._hidden || force_show || (force_show==null && src.auto_show))
			src.Show(force_timeout) // Show the statbar


	proc/Show(force_timeout=null) // Show the statbar

		if(!src.owner)
			CRASH("bar_overlay: owner for '[src.var_stat]/[src.var_max]' bar no longer exists!")

		if(!src._image)
			// Let Update() handle the call and generate an image for us
			Update(null, TRUE, force_timeout)
			return

		src._timeout_id++ // Increment the timeout id to cancel any running timeouts

		if(force_timeout == null)
			force_timeout = src.timeout // Default to src.timeout if none was passed

		src._hidden=0 // Mark statbar as not hidden

		// In case the image is already present, we always remove it first to prevent duplicates
		owner.overlays -= src._image

		// Add the bar to the overlays
		owner.overlays += src._image

		if(force_timeout) // If the bar should auto-hide

			var/current_id = src._timeout_id // Store the current id

			spawn(force_timeout) // Sleep for the ticks specified

				if(src._timeout_id == current_id) // If another Show() hasn't been called since

					if(src.owner) // If the owner hasn't been deleted

						src.Hide() // Hide the statbar

	proc/Hide() // Hide the statbar

		if(!src.owner)
			CRASH("bar_overlay: owner for '[src.var_stat]/[src.var_max]' bar no longer exists!")

		src._timeout_id++ // Increment the timeout id to cancel any running timeouts

		src._hidden=1 // Mark statbar as hidden

		if(src._image)
			// Remove statbar image
			owner.overlays -= src._image